Package: batchelor
Version: 1.28.0
Date: 2026-04-09
Title: Single-Cell Batch Correction Methods
Authors@R: c(person("Aaron", "Lun", role = c("aut", "cre"), email = "infinite.monkeys.with.keyboards@gmail.com"),
    person("Laleh", "Haghverdi", role="ctb"))
Depends: SingleCellExperiment
Imports: SummarizedExperiment, S4Vectors, BiocGenerics, Rcpp, stats,
        methods, utils, igraph, BiocNeighbors, BiocSingular, Matrix,
        SparseArray, DelayedArray (>= 0.31.5), DelayedMatrixStats,
        BiocParallel, scuttle, ResidualMatrix, ScaledMatrix, beachmat
Suggests: testthat, BiocStyle, knitr, rmarkdown, scran, scater,
        bluster, scRNAseq
biocViews: Sequencing, RNASeq, Software, GeneExpression,
        Transcriptomics, SingleCell, BatchEffect, Normalization
LinkingTo: Rcpp
Description: 
    Implements a variety of methods for batch correction of single-cell (RNA sequencing) data. 
    This includes methods based on detecting mutually nearest neighbors,
    as well as several efficient variants of linear regression of the log-expression values.
    Functions are also provided to perform global rescaling to remove differences in depth between batches,
    and to perform a principal components analysis that is robust to differences in the numbers of cells across batches.
License: GPL-3
NeedsCompilation: yes
VignetteBuilder: knitr
SystemRequirements: C++11
RoxygenNote: 7.3.2
git_url: https://git.bioconductor.org/packages/batchelor
git_branch: RELEASE_3_23
git_last_commit: 5823437
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.23
Date/Publication: 2026-04-28
Packaged: 2026-04-28 21:29:23 UTC; biocbuild
Author: Aaron Lun [aut, cre],
  Laleh Haghverdi [ctb]
Maintainer: Aaron Lun <infinite.monkeys.with.keyboards@gmail.com>
Built: R 4.6.0; aarch64-apple-darwin23; 2026-04-29 04:34:04 UTC; unix
