# Generated by roxygen2: do not edit by hand

export(ClassicMnnParam)
export(FastMnnParam)
export(NoCorrectParam)
export(RegressParam)
export(RescaleParam)
export(applyMultiSCE)
export(batchCorrect)
export(checkBatchConsistency)
export(checkIfSCE)
export(checkRestrictions)
export(clusterAbundanceTest)
export(clusterAbundanceVar)
export(clusterMNN)
export(convertPCsToSCE)
export(correctExperiments)
export(cosineNorm)
export(divideIntoBatches)
export(fastMNN)
export(findMutualNN)
export(intersectRows)
export(mnnCorrect)
export(mnnDeltaVariance)
export(multiBatchNorm)
export(multiBatchPCA)
export(noCorrect)
export(quickCorrect)
export(reducedMNN)
export(regressBatches)
export(rescaleBatches)
exportClasses(BatchelorParam)
exportClasses(ClassicMnnParam)
exportClasses(FastMnnParam)
exportClasses(NoCorrectParam)
exportClasses(RegressParam)
exportClasses(RescaleParam)
exportMethods(batchCorrect)
import(SparseArray)
import(methods)
importClassesFrom(DelayedArray,integer_OR_NULL)
importClassesFrom(Matrix,dsparseMatrix)
importClassesFrom(S4Vectors,DataFrame)
importClassesFrom(S4Vectors,List)
importClassesFrom(S4Vectors,SimpleList)
importClassesFrom(SingleCellExperiment,SingleCellExperiment)
importFrom(BiocGenerics,"colnames<-")
importFrom(BiocGenerics,"rownames<-")
importFrom(BiocGenerics,"sizeFactors<-")
importFrom(BiocGenerics,cbind)
importFrom(BiocGenerics,colnames)
importFrom(BiocGenerics,ncol)
importFrom(BiocGenerics,rbind)
importFrom(BiocGenerics,rownames)
importFrom(BiocGenerics,sizeFactors)
importFrom(BiocNeighbors,KmknnParam)
importFrom(BiocNeighbors,findMutualNN)
importFrom(BiocNeighbors,queryKNN)
importFrom(BiocParallel,SerialParam)
importFrom(BiocParallel,bpstart)
importFrom(BiocParallel,bpstop)
importFrom(BiocSingular,ExactParam)
importFrom(BiocSingular,IrlbaParam)
importFrom(BiocSingular,LowRankMatrix)
importFrom(BiocSingular,bsdeferred)
importFrom(BiocSingular,runPCA)
importFrom(BiocSingular,runSVD)
importFrom(DelayedArray,"seed<-")
importFrom(DelayedArray,DelayedArray)
importFrom(DelayedArray,blockApply)
importFrom(DelayedArray,getAutoBPPARAM)
importFrom(DelayedArray,rowAutoGrid)
importFrom(DelayedArray,seed)
importFrom(DelayedArray,setAutoBPPARAM)
importFrom(DelayedArray,sweep)
importFrom(DelayedMatrixStats,colVars)
importFrom(DelayedMatrixStats,rowVars)
importFrom(Matrix,Diagonal)
importFrom(Matrix,colSums)
importFrom(Matrix,crossprod)
importFrom(Matrix,rowMeans)
importFrom(Matrix,rowSums)
importFrom(Matrix,sparseMatrix)
importFrom(Matrix,t)
importFrom(Rcpp,sourceCpp)
importFrom(ResidualMatrix,ResidualMatrix)
importFrom(S4Vectors,"mcols<-")
importFrom(S4Vectors,"metadata<-")
importFrom(S4Vectors,DataFrame)
importFrom(S4Vectors,I)
importFrom(S4Vectors,List)
importFrom(S4Vectors,SimpleList)
importFrom(S4Vectors,make_zero_col_DFrame)
importFrom(S4Vectors,mcols)
importFrom(S4Vectors,metadata)
importFrom(S4Vectors,normalizeSingleBracketSubscript)
importFrom(ScaledMatrix,ScaledMatrix)
importFrom(SingleCellExperiment,"altExp<-")
importFrom(SingleCellExperiment,"reducedDim<-")
importFrom(SingleCellExperiment,"reducedDims<-")
importFrom(SingleCellExperiment,SingleCellExperiment)
importFrom(SingleCellExperiment,altExp)
importFrom(SingleCellExperiment,altExpNames)
importFrom(SingleCellExperiment,reducedDim)
importFrom(SingleCellExperiment,reducedDimNames)
importFrom(SingleCellExperiment,reducedDims)
importFrom(SingleCellExperiment,simplifyToSCE)
importFrom(SummarizedExperiment,"assay<-")
importFrom(SummarizedExperiment,"assays<-")
importFrom(SummarizedExperiment,"colData<-")
importFrom(SummarizedExperiment,"rowData<-")
importFrom(SummarizedExperiment,"rowRanges<-")
importFrom(SummarizedExperiment,assay)
importFrom(SummarizedExperiment,assayNames)
importFrom(SummarizedExperiment,assays)
importFrom(SummarizedExperiment,colData)
importFrom(SummarizedExperiment,rowData)
importFrom(SummarizedExperiment,rowRanges)
importFrom(beachmat,realizeFileBackedMatrix)
importFrom(igraph,components)
importFrom(igraph,make_graph)
importFrom(methods,as)
importFrom(methods,is)
importFrom(methods,new)
importFrom(scuttle,.bpNotSharedOrUp)
importFrom(scuttle,.subset2index)
importFrom(scuttle,.unpackLists)
importFrom(scuttle,calculateAverage)
importFrom(scuttle,librarySizeFactors)
importFrom(scuttle,logNormCounts)
importFrom(scuttle,normalizeCounts)
importFrom(scuttle,sumCountsAcrossCells)
importFrom(stats,chisq.test)
importFrom(stats,median)
importFrom(stats,model.matrix)
importFrom(utils,head)
importFrom(utils,relist)
importFrom(utils,tail)
importMethodsFrom(BiocGenerics,ncol)
importMethodsFrom(BiocGenerics,nrow)
useDynLib(batchelor)
