# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

adjust_shift_variance <- function(data1, data2, vect, sigma2, restrict1, restrict2) {
    .Call('_batchelor_adjust_shift_variance', PACKAGE = 'batchelor', data1, data2, vect, sigma2, restrict1, restrict2)
}

find_mutual_nns <- function(left, right) {
    .Call('_batchelor_find_mutual_nns', PACKAGE = 'batchelor', left, right)
}

smooth_gaussian_kernel <- function(averaged, index, mat, sigma2) {
    .Call('_batchelor_smooth_gaussian_kernel', PACKAGE = 'batchelor', averaged, index, mat, sigma2)
}

